import numpy as np
import pygrib as pg
import os, sys
import json
import pprint
import tensorflow as tf
import matplotlib.pyplot as plt
import glob
from datetime import datetime, timedelta

import inspect

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.insert(0, os.path.dirname(current_dir))

from lightningcast import utils
from lightningcast import ltg_utils
import netCDF4
from pyproj import Proj
import time
import pyresample as pr
import logging
from lightningcast.sat_zen_angle import sat_zen_angle
from lightningcast.solar_angles import solar_angles

os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
pltg = os.environ["PLTG"]
# ------------------------------------------------------------------------------------------
def get_remap_cache(input_def, target_def, radius_of_influence=4000):
    logging.info("Making remap cache...")
    (
        input_index,
        output_index,
        index_array,
        distance_array,
    ) = pr.kd_tree.get_neighbour_info(
        input_def, target_def, radius_of_influence, neighbours=1
    )
    remap_cache = {
        "input_index": input_index,
        "output_index": output_index,
        "index_array": index_array,
        "target_def": target_def,
    }
    return remap_cache


# ------------------------------------------------------------------------------------------
def image_feature(value):
    """Returns a bytes_list from a string / byte."""
    # return tf.train.Feature(
    #    bytes_list=tf.train.BytesList(value=[tf.io.encode_jpeg(value).numpy()])
    # )
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value.numpy()]))


def bytes_feature(value):
    """Returns a bytes_list from a string / byte."""
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value.encode()]))


def float_feature(value):
    """Returns a float_list from a float / double."""
    return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))


def int64_feature(value):
    """Returns an int64_list from a bool / enum / int / uint."""
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def float_feature_list(value):
    """Returns a list of float_list from a float / double."""
    return tf.train.Feature(float_list=tf.train.FloatList(value=value))


def create_example(channels):
    feature = {}
    for chan in channels:
        feature[chan] = image_feature(channels[chan])
    return tf.train.Example(features=tf.train.Features(feature=feature))


# ---------------------------------------------------------------------------------------------------------------
def get_satzen(georeference_file):
    nc = netCDF4.Dataset(georeference_file, "r")
    gip = nc.variables["goes_imager_projection"]
    x = nc.variables["x"][:]
    y = nc.variables["y"][:]
    projection = Proj(
        f"+proj=geos +lon_0={gip.longitude_of_projection_origin}"
        + f" +h={gip.perspective_point_height} +x_0=0.0 +y_0=0.0 +a="
        + f"{gip.semi_major_axis} +b={gip.semi_minor_axis}"
    )
    x *= gip.perspective_point_height
    y *= gip.perspective_point_height
    xx, yy = np.meshgrid(x, y)
    lons2D, lats2D = projection(xx, yy, inverse=True)
    satzen2D = sat_zen_angle(
        xlat=lats2D, xlon=lons2D, satlon=gip.longitude_of_projection_origin
    )
    return np.nan_to_num(satzen2D, nan=255), lats2D, lons2D


# ---------------------------------------------------------------------------------------------------------------
def find_mrms_data(abidt, radpatt, radgrid, goes_area_def, bsinfo):

    for mmm in [0, 1, 2, -1, -2, -3, 3]:
        raddt = abidt + timedelta(minutes=mmm)
        try:
            radfile = glob.glob(raddt.strftime(radpatt))[0]
        except IndexError:
            if mmm == 3:
                print(f"WARNING: Can't find ref10C for sat dt: {dt}")
                raise
        else:
            # Now get the radar data
            grb = pg.open(radfile)
            msg = grb[1]  # should only be 1 'message' in grib2 file
            data = msg.values
            grb.close()
            rad_remapped = pr.kd_tree.resample_nearest(
                radgrid, data, goes_area_def, radius_of_influence=4000, fill_value=-999
            )
            # Special bytescaling for Ref10 to account for "on grid" but missing
            rad_rescaled = (
                utils.bytescale(
                    rad_remapped, bsinfo["REF10"]["vmin"], bsinfo["REF10"]["vmax"]
                )
                + 10
            )
            rad_rescaled[rad_remapped <= -999] = 0
            rad_rescaled[rad_remapped == -99] = 1
            rad_rescaled[rad_rescaled > 255] = 255
            rad_rescaled = rad_rescaled.astype(np.uint8)
            rad_rescaled = np.expand_dims(rad_rescaled, axis=-1)
            break

    return rad_rescaled


# ---------------------------------------------------------------------------------------------------------------
bsinfo = {}  # bytescaling info
bsinfo["CH01"] = {"vmin": 0, "vmax": 1}
bsinfo["CH02"] = {"vmin": 0, "vmax": 1}
bsinfo["CH03"] = {"vmin": 0, "vmax": 1}
bsinfo["CH04"] = {"vmin": 0, "vmax": 0.75}
bsinfo["CH05"] = {"vmin": 0, "vmax": 0.75}
bsinfo["CH06"] = {"vmin": 0, "vmax": 0.75}
bsinfo["CH07"] = {"vmin": 220, "vmax": 360}
bsinfo["CH08"] = {"vmin": 200, "vmax": 270}
bsinfo["CH09"] = {"vmin": 200, "vmax": 275}
bsinfo["CH10"] = {"vmin": 200, "vmax": 280}
bsinfo["CH11"] = {"vmin": 190, "vmax": 320}
bsinfo["CH12"] = {"vmin": 190, "vmax": 320}
bsinfo["CH13"] = {"vmin": 190, "vmax": 320}
bsinfo["CH14"] = {"vmin": 190, "vmax": 320}
bsinfo["CH15"] = {"vmin": 190, "vmax": 320}
bsinfo["CH16"] = {"vmin": 190, "vmax": 300}
bsinfo["FED"] = {"vmin": 0, "vmax": 255}
bsinfo["FED_accum_60min_2km"] = {"vmin": 0, "vmax": 255}
bsinfo["FED_accum_15min_2km"] = {"vmin": 0, "vmax": 255}
bsinfo["satzen"] = {"vmin": 0, "vmax": 90}
bsinfo["REF10"] = {"vmin": 0, "vmax": 75}

# mrms info
var = "Reflectivity_-10C"
radpatt = f"/ships22/grain/probsevere/radar/%Y/%Y%m%d/grib2/{var}/MRMS_{var}_00.50_%Y%m%d-%H%M*.grib2"
NWlat, NWlon = (55, -130)
nlats = 3500
nlons = 7000
radlats = np.arange(55, 20, -0.01)
radlons = np.arange(-130, -60, 0.01)
radlons, radlats = np.meshgrid(radlons, radlats)
radgrid = pr.geometry.GridDefinition(lons=radlons, lats=radlats)

# goes info
sat = "goes18"
position = "goes_east" if (sat == "goes16") else "goes_west"
abipatt = f"/arcdata/goes/grb/{sat}/%Y/%Y_%m_%d_%j/abi/L1b/RadC/"
glmdir = f"/ships19/grain/probsevere/lightning/%Y/%Y%m%d/GLM/{position}/"
glmpatt1 = f"{glmdir}/FED_accum_60min_2km/%Y%m%d-%H%M00.netcdf"
glmvar1 = "FED_accum_60min_2km"
# glmpatt2 = f'{glmdir}/FED_accum_15min_2km/%Y%m%d-%H%M00.netcdf'; glmvar2 = "FED_accum_15min_2km"
aggpatt = f"{glmdir}/agg/%Y%m%d-%H%M00.netcdf"
sector = "RadC"
ltgthresh = 3  # should be in byte-scaled space

process_ref10 = False

startdt = dt = datetime(2023, 9, 15, 0, 0)
enddt = datetime(2023, 12, 31, 23, 59)

outpatt = f"/ships22/grain/probsevere/LC/tfrecs3/{sat}/%Y/%Y%m%d/"

assert startdt <= enddt

# For 2-km data
# ny,nx = (736,832)
ny, nx = (480, 480)
NorthwestXs = np.arange(50, 2500 - nx, nx, dtype=int)
NorthwestYs = np.arange(50, 1500 - ny, ny, dtype=int)

if sat == "goes16":  # GOES_East GLM and ABI are exactly the same!
    glm_area_def, _, _ = utils.get_area_definition(
        f"{pltg}/lightningcast/static/GOES_East.nc"
    )
    abi_area_def = glm_area_def
    goes_1km_def, _, _ = utils.get_area_definition(
        "/arcdata/goes/grb/goes16/2023/2023_06_30_181/abi/L1b/RadC/OR_ABI-L1b-RadC-M6C05_G16_s20231811801178_e20231811803551_c20231811803597.nc"
    )
    satzen2D, lats, lons = get_satzen(f"{pltg}/lightningcast/static/GOES_East.nc")
    satzen2D = utils.bytescale(
        satzen2D, bsinfo["satzen"]["vmin"], bsinfo["satzen"]["vmax"]
    )
    remap_cache = None  # G16 GLM and ABI on same projection
elif sat == "goes18":  # GOES_West GLM and ABI are slightly different, for some reason.
    glm_area_def, _, _ = utils.get_area_definition(
        f"{pltg}/lightningcast/static/GOES_West_GLM.nc"
    )
    abi_area_def, _, _ = utils.get_area_definition(
        f"{pltg}/lightningcast/static/GOES_West.nc"
    )
    remap_cache = get_remap_cache(glm_area_def, abi_area_def)


while startdt <= dt <= enddt:

    glmdt = dt + timedelta(minutes=60)
    glmfile = glob.glob(glmdt.strftime(glmpatt1))
    if len(glmfile):
        try:
            if (
                sat == "goes16"
            ):  # For G16 CONUS, FED returned as masked array. For G18 PACUS, returned as regular np.ndarray (min=0).
                fedaccum60 = ltg_utils.get_fed(
                    glmdt, glmpatt1, glmvar=glmvar1, remap_cache=remap_cache
                ).filled(fill_value=0)
            else:
                fedaccum60 = ltg_utils.get_fed(
                    glmdt, glmpatt1, glmvar=glmvar1, remap_cache=remap_cache
                )

            fedaccum60 = np.expand_dims(
                utils.bytescale(
                    fedaccum60, bsinfo[glmvar1]["vmin"], bsinfo[glmvar1]["vmax"]
                ),
                axis=-1,
            )

            # fedaccum15 = ltg_utils.get_fed(glmdt,glmpatt2,glmvar=glmvar2,remap_cache=remap_cache).filled(fill_value=0)
            # fedaccum15 = np.expand_dims(bytescale(fedaccum15,bsinfo[glmvar2]['vmin'],bsinfo[glmvar2]['vmax']), axis=-1

            # I guess the 5-min agg FED files are always masked arrays?
            fed = ltg_utils.get_fed(dt, aggpatt, glmvar="flash_extent_density").filled(
                fill_value=0
            )  # ,remap_cache=remap_cache)
            fed = np.expand_dims(
                utils.bytescale(fed, bsinfo["FED"]["vmin"], bsinfo["FED"]["vmax"]),
                axis=-1,
            )
        except (UnboundLocalError, IndexError) as err:
            pass
        else:
            if np.max(fedaccum60) > ltgthresh:
                # Collect all of the data, convert to np.ndarray, bytescale, and put in tensor form (still np.ndarray)

                try:
                    if process_ref10:
                        ref10 = find_mrms_data(
                            dt, radpatt, radgrid, goes_1km_def, bsinfo
                        )
                    else:
                        _ = 0
                except IndexError:
                    pass
                else:
                    try:
                        ch02 = utils.get_abi_data(dt, "02", abipatt).filled(
                            fill_value=0
                        )
                        ch02 = np.expand_dims(
                            utils.bytescale(
                                ch02, bsinfo["CH02"]["vmin"], bsinfo["CH02"]["vmax"]
                            ),
                            axis=-1,
                        )

                        ch05 = utils.get_abi_data(dt, "05", abipatt).filled(
                            fill_value=0
                        )
                        ch05 = np.expand_dims(
                            utils.bytescale(
                                ch05, bsinfo["CH05"]["vmin"], bsinfo["CH05"]["vmax"]
                            ),
                            axis=-1,
                        )

                        ch08 = utils.get_abi_data(dt, "08", abipatt).filled(
                            fill_value=0
                        )
                        ch08 = np.expand_dims(
                            utils.bytescale(
                                ch08, bsinfo["CH08"]["vmin"], bsinfo["CH08"]["vmax"]
                            ),
                            axis=-1,
                        )

                        ch09 = utils.get_abi_data(dt, "09", abipatt).filled(
                            fill_value=0
                        )
                        ch09 = np.expand_dims(
                            utils.bytescale(
                                ch09, bsinfo["CH09"]["vmin"], bsinfo["CH09"]["vmax"]
                            ),
                            axis=-1,
                        )

                        ch10 = utils.get_abi_data(dt, "10", abipatt).filled(
                            fill_value=0
                        )
                        ch10 = np.expand_dims(
                            utils.bytescale(
                                ch10, bsinfo["CH10"]["vmin"], bsinfo["CH10"]["vmax"]
                            ),
                            axis=-1,
                        )

                        ch11 = utils.get_abi_data(dt, "11", abipatt).filled(
                            fill_value=0
                        )
                        ch11 = np.expand_dims(
                            utils.bytescale(
                                ch11, bsinfo["CH11"]["vmin"], bsinfo["CH11"]["vmax"]
                            ),
                            axis=-1,
                        )

                        ch13 = utils.get_abi_data(dt, "13", abipatt).filled(
                            fill_value=0
                        )
                        ch13 = np.expand_dims(
                            utils.bytescale(
                                ch13, bsinfo["CH13"]["vmin"], bsinfo["CH13"]["vmax"]
                            ),
                            axis=-1,
                        )

                        ch15 = utils.get_abi_data(dt, "15", abipatt).filled(
                            fill_value=0
                        )
                        ch15 = np.expand_dims(
                            utils.bytescale(
                                ch15, bsinfo["CH15"]["vmin"], bsinfo["CH15"]["vmax"]
                            ),
                            axis=-1,
                        )
                    except (UnboundLocalError, IndexError) as err:
                        print(err)
                    else:
                        print(dt)

                        # Create patches, convert patches to TF tensors, and serialize
                        for Y in NorthwestYs:
                            for X in NorthwestXs:
                                if (
                                    np.max(fedaccum60[Y : Y + ny, X : X + nx])
                                    >= ltgthresh
                                ):
                                    outdir = dt.strftime(outpatt)
                                    utils.mkdir_p(outdir)
                                    outfile = f"{outdir}/{sat}_{dt.strftime('%Y%m%d-%H%M%S')}_Y{Y}_X{X}.tfrec"

                                    with tf.io.TFRecordWriter(outfile) as writer:
                                        channels = {}
                                        t1 = time.time()
                                        channels["CH02"] = tf.io.serialize_tensor(
                                            tf.convert_to_tensor(
                                                ch02[
                                                    Y * 4 : (Y + ny) * 4,
                                                    X * 4 : (X + nx) * 4,
                                                ],
                                                dtype=tf.uint8,
                                            )
                                        )
                                        channels["CH05"] = tf.io.serialize_tensor(
                                            tf.convert_to_tensor(
                                                ch05[
                                                    Y * 2 : (Y + ny) * 2,
                                                    X * 2 : (X + nx) * 2,
                                                ],
                                                dtype=tf.uint8,
                                            )
                                        )
                                        channels["CH08"] = tf.io.serialize_tensor(
                                            tf.convert_to_tensor(
                                                ch08[Y : Y + ny, X : X + nx],
                                                dtype=tf.uint8,
                                            )
                                        )
                                        channels["CH09"] = tf.io.serialize_tensor(
                                            tf.convert_to_tensor(
                                                ch09[Y : Y + ny, X : X + nx],
                                                dtype=tf.uint8,
                                            )
                                        )
                                        channels["CH10"] = tf.io.serialize_tensor(
                                            tf.convert_to_tensor(
                                                ch10[Y : Y + ny, X : X + nx],
                                                dtype=tf.uint8,
                                            )
                                        )
                                        channels["CH11"] = tf.io.serialize_tensor(
                                            tf.convert_to_tensor(
                                                ch11[Y : Y + ny, X : X + nx],
                                                dtype=tf.uint8,
                                            )
                                        )
                                        channels["CH13"] = tf.io.serialize_tensor(
                                            tf.convert_to_tensor(
                                                ch13[Y : Y + ny, X : X + nx],
                                                dtype=tf.uint8,
                                            )
                                        )
                                        channels["CH15"] = tf.io.serialize_tensor(
                                            tf.convert_to_tensor(
                                                ch15[Y : Y + ny, X : X + nx],
                                                dtype=tf.uint8,
                                            )
                                        )
                                        channels["FED"] = tf.io.serialize_tensor(
                                            tf.convert_to_tensor(
                                                fed[Y : Y + ny, X : X + nx],
                                                dtype=tf.uint8,
                                            )
                                        )
                                        channels[
                                            "FED_accum_60min_2km"
                                        ] = tf.io.serialize_tensor(
                                            tf.convert_to_tensor(
                                                fedaccum60[Y : Y + ny, X : X + nx],
                                                dtype=tf.uint8,
                                            )
                                        )

                                        if process_ref10:
                                            channels["REF10"] = tf.io.serialize_tensor(
                                                tf.convert_to_tensor(
                                                    ref10[
                                                        Y * 2 : (Y + ny) * 2,
                                                        X * 2 : (X + nx) * 2,
                                                    ],
                                                    dtype=tf.uint8,
                                                )
                                            )
                                        # channels['FED_accum_15min_2km'] = tf.io.serialize_tensor(tf.convert_to_tensor(fedaccum15[Y:Y+ny,X:X+nx], dtype=tf.uint8))
                                        # channels['SATZEN'] = tf.io.serialize_tensor(tf.convert_to_tensor(satzen2D[Y:Y+ny,X:X+nx], dtype=tf.uint8))
                                        example = create_example(channels)
                                        writer.write(example.SerializeToString())

                    # If we had a successful set of writes, add 6 min. Then subsequently add 1 min until
                    # we get to the next file. In this way, we are essentially doing every 10 minutes.
                    dt += timedelta(minutes=6)

    dt += timedelta(minutes=1)
