import sys
from glob import glob
import subprocess
import os
import numpy as np
import json
import argparse
from datetime import datetime, timedelta
import pickle
import inspect

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.insert(0, os.path.dirname(current_dir))
from lightningcast.__version__ import __version__

pltg_data = os.environ["PLTG_DATA"]
##########################################################################################################################################
def pltg_gr_placefile():

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-st",
        "--start_time",
        help="If using the TimeRange capability (realtime or offline). Format='20170420-175521'",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-et",
        "--end_time",
        help="If using the TimeRange capability (realtime or offline). Format='20170420-185521'",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-d",
        "--datapatt",
        help="Directory pattern needed for offline TimeRange file. "
        + "Default = ${PLTG_DATA}/products/%%Y%%m%%d/goes_east/RadC/json_plax/1fl-60min/",
        default=[f"{pltg_data}/products/%Y%m%d/goes_east/RadC/json_plax/1fl-60min/"],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-j",
        "--json_file",
        help="full file path and name. One time only.",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-o",
        "--outdirpatt",
        help="output pattern dir for files. Default = ${PLTG_DATA}",
        default=[pltg_data],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-of",
        "--outfile",
        help="outfile name for PLTG placefile. Default='NOAACIMSS_PLTG_{satellite}_{sector}_LOOP' (if CIMSS)",
        default=[None],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-r",
        "--realtime",
        help="If realtime, files will be uploaded to web sever and have a static filename.",
        action="store_true",
    )

    args = parser.parse_args()


    if args.json_file:
        singleGRFile(
            args.json_file[0], outdirpatt[0], outfile=args.outfile[0], realtime=args.realtime
        )
    elif args.start_time and args.end_time:
        timeRangeGRFile(
            args.start_time[0],
            args.end_time[0],
            args.datapatt[0],
            outfile=args.outfile[0],
            outdirpatt=args.outdirpatt[0],
            realtime=args.realtime,
        )


#########################################################################################################################################
def timeRangeGRFile(
    start_timestamp, end_timestamp, datapatt, outdirpatt, outfile=None, realtime=False
):

    start_datetime = datetime.strptime(start_timestamp, "%Y%m%d-%H%M%S")
    start_date = start_datetime.strftime("%Y%m%d")
    start_epoch = int(start_datetime.strftime("%s"))
    end_datetime = datetime.strptime(end_timestamp, "%Y%m%d-%H%M%S")
    end_date = end_datetime.strftime("%Y%m%d")
    end_epoch = int(end_datetime.strftime("%s"))

    outdir = end_datetime.strftime(outdirpatt)  # most recent
    os.makedirs(outdir, exist_ok=True)

    prefix = "LtgCast"
    if start_date == end_date:
        jsons = glob(start_datetime.strftime(f"{datapatt}/{prefix}*.json"))
    else:  # only works across 2 dates
        jsons = glob(start_datetime.strftime(f"{datapatt}/{prefix}*.json")) + glob(
            end_datetime.strftime(f"{datapatt}/{prefix}*.json")
        )
    jsons = np.sort(jsons)

    with open(jsons[0], "r") as f:
        data = json.load(f)
        sector = data["sector"]
        slot = data["orbital_slot"]

    if sector == "RadC":
        sector = "CONUS"
        NCCF_sector = "RadC"
    elif sector == "RadM1":
        sector = "Meso1"
        NCCF_sector = "RadM1"
    elif sector == "RadM2":
        sector = "Meso2"
        NCCF_sector = "RadM2"
    else:
        print("Placefiles only supported for RadC, RadM1, and RadM2")
        sys.exit(1)

    if slot != "GOES-East" and slot != "GOES-West":
        print("Placefiles only supported for GOES-East and GOES-West.")
        sys.exit(1)

    # Looping LightningCast
    startTimes = []
    good_jsons = []
    good_epochs = []  # list of epochs seconds for validtimes
    for json_file in jsons:
        parts = os.path.basename(json_file).split("_s")
        validDateTime = parts[1][0:14]
        tmpDT = datetime.strptime(f"{validDateTime}", "%Y%m%d%H%M%S")
        tmpEpochSecs = int(tmpDT.strftime("%s"))
        if tmpEpochSecs >= start_epoch and tmpEpochSecs <= end_epoch:
            startTimes.append(
                tmpDT
            )  # .append(f"{validDate[0:4]}-{validDate[4:6]}-{validDate[6:8]}T{validTime[0:2]}:{validTime[2:4]}:{validTime[4:6]}Z")
            good_jsons.append(json_file)
            good_epochs.append(tmpEpochSecs)
        else:
            continue  # outside our time range

    if len(startTimes) > 1:
        endTimes = startTimes[1:]
        # get last endTime
        last_endTime = endTimes[-1] + timedelta(minutes=10)
        endTimes.append(last_endTime)
    else:
        endTimes = [startTimes[0] + timedelta(minutes=10)]

    assert len(good_jsons) > 0

    if realtime:  # i.e., infinite processing at CIMSS. Use static filenames
        outfile = f"NOAACIMSS_PLTG_{slot}_{sector}_LOOP"
    else:
        if outfile is None:
            # Output filename for not "realtime" (i.e., not CIMSS)
            # Get sat_abbr from first good_json
            with open(good_jsons[0]) as json_data:
                data = json.load(json_data)
            sat_abbr = data["platform"][0].lower() + data["platform"].split("-")[1]
            assert sat_abbr[0] == "g"
            outfile = (
                f"LtgCast-pc-{NCCF_sector}_v{__version__.replace('.','-')}_{sat_abbr}"
            )
            outfile += f"_s{start_datetime.strftime('%Y%m%d%H%M%S%f')[0:15]}"
            outfile += f"_e{end_datetime.strftime('%Y%m%d%H%M%S%f')[0:15]}"
            outfile += (
                f"_c{datetime.utcnow().strftime('%Y%m%d%H%M%S%f')[0:15]}.placefile"
            )

    ps = open(f"{outdir}/{outfile}", "w")
    ps.write(f"Title: NOAA/CIMSS LightningCast loop -- {slot} {sector} sector\n")
    ps.write("Refresh: 1\n\n")
    for i in range(len(good_jsons)):
        with open(good_jsons[i]) as json_data:
            data = json.load(json_data)

        timestamp = datetime.utcfromtimestamp(good_epochs[i]).strftime(
            "%Y-%m-%d %H:%MZ"
        )
        ps.write(
            f"TimeRange: {startTimes[i].strftime('%Y-%m-%dT%H:%M:%SZ')} {endTimes[i].strftime('%Y-%m-%dT%H:%M:%SZ')}\n\n"
        )

        for storm in data["features"]:
            props = storm["properties"]
            coord_str = ""
            try:
                coords = storm["geometry"]["coordinates"][0]
                for coord in coords:
                    coord_str += f"{coord[1]}, {coord[0]}\n"
            except TypeError:  # indicates that it's a LineString instead of Polygon
                coords = storm["geometry"]["coordinates"]
                for coord in coords:
                    coord_str += f"{coord[1]}, {coord[0]}\n"

            rgb = (" ").join([ss.zfill(3) for ss in props["COLOR"].split()])
            bwidth = "2"

            prob = int(props["INFO"].split("):")[1].split("%")[0])
            txt_lines = [f"{slot} {sector} scan time: {timestamp}\\n"]
            txt_lines.append(
                f"P(LTG) >= {prob}% in next 60 minutes at every encompassed location."
            )

            hover_str = ("").join(txt_lines)

            ps.write(f"Color: {rgb} \n")
            ps.write(f'Line: {bwidth}, 0, "{hover_str}"\n')
            ps.write(coord_str)
            ps.write("End:\n\n")

    ps.close()
    print(f"Wrote out {outdir}/{outfile}")

    if realtime:
        try:
            s = subprocess.run(
                [
                    "rsync",
                    "-avuz",
                    f"{outdir}/{outfile}",
                    "webaccess:/webdata/web/cimss/htdocs/severe_conv/",
                ],
                timeout=10,
            )
        except subprocess.TimeoutExpired as err:
            logging.error(str(err))


######################################################################################################################################
def singleGRFile(json_file, outdir, outfile=None, realtime=False):

    basefile = os.path.basename(json_file)

    with open(json_file) as json_data:
        data = json.load(json_data)
    parts = basefile.split("_")
    validDate = parts[1]
    validTime = parts[2][0:6]
    dt = datetime.strptime(validDate + validTime, "%Y%m%d%H%M%S")
    timestamp = dt.strftime("%Y-%m-%d %H:%MZ")

    if "goes_east" in json_file:
        slot = "GOES-East"
    elif "goes_west" in json_file:
        slot = "GOES-West"
    else:
        print("Placefiles only supported for GOES-East and GOES-West.")
        sys.exit(1)
    if "RadC" in json_file:
        sector = "CONUS"
    elif "RadM1" in json_file:
        sector = "Meso1"
    elif "RadM2" in json_file:
        sector = "Meso2"
    else:
        print("Placefiles only supported for RadC, RadM1, and RadM2")
        sys.exit(1)
    if outfile is None:
        outfile = f"NOAACIMSS_PLTG_{slot}_{sector}"

    if not (realtime):
        outfile += f"_{validDate}-{validTime}"

    ps = open(f"{outdir}/{outfile}", "w")
    ps.write(f"Title: NOAA/CIMSS LightningCast loop -- {slot} {sector} sector\n")
    ps.write("Refresh: 1\n\n")

    for storm in data["features"]:
        props = storm["properties"]
        coord_str = ""
        try:
            coords = storm["geometry"]["coordinates"][0]
            for coord in coords:
                coord_str += str(coord[1]) + ", " + str(coord[0]) + "\n"
        except TypeError:  # indicates that it's a LineString instead of Polygon
            coords = storm["geometry"]["coordinates"]
            for coord in coords:
                coord_str += f"{coord[1]}, {coord[0]}\n"

        rgb = (" ").join([ss.zfill(3) for ss in props["COLOR"].split()])
        bwidth = "2"

        prob = int(props["INFO"].split()[1].split("%")[0])
        txt_lines = [f"{slot} {sector} scan time: {timestamp}\\n"]
        txt_lines.append(
            f"P(LTG) ≥ {prob}% in next 60 minutes at any encompassed location."
        )

        hover_str = ("").join(txt_lines)

        ps.write(f"Color: {rgb} \n")
        ps.write(f'Line: {bwidth}, 0, "{hover_str}"\n')
        ps.write(coord_str)
        ps.write("End:\n\n")

    ps.close()
    print(f"Wrote out {outdir}/{outfile}")

    if realtime:
        try:
            s = subprocess.run(
                [
                    "rsync",
                    "-avuz",
                    f"{outdir}/{outfile}",
                    "webaccess:/webdata/web/cimss/htdocs/severe_conv/",
                ],
                timeout=10,
            )
        except subprocess.TimeoutExpired as err:
            logging.error(str(err))


if __name__ == "__main__":
    pltg_gr_placefile()
