from __future__ import absolute_import, division, print_function, unicode_literals

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import os, sys

pltg_data = os.environ["PLTG_DATA"] + "/"
pltg = os.environ["PLTG"] + "/"
sys.path.insert(0, os.path.join(pltg, "src", "static/"))
from scipy.ndimage import zoom

# os.environ['CUDA_DEVICE_ORDER']="PCI_BUS_ID"
# os.environ["CUDA_VISIBLE_DEVICES"]=""
from netCDF4 import Dataset
import numpy as np
import argparse
import pandas as pd

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import Model, save_model, load_model

from train_eval_cnn import normalize_multivariate_data as nmd
from lightningcast import keras_metrics
import matplotlib.pyplot as plt
import time
from subprocess import PIPE, call, Popen
from pyproj import Proj
from ABIcmaps import irw_ctc
from lightningcast import utils
import numpy.ma as ma
from datetime import datetime, timedelta, timezone
import re
from lightningcast.solar_angles import solar_angles
from lightningcast.sat_zen_angle import sat_zen_angle
import pickle
import glob
from matplotlib.lines import Line2D
import shutil
from collections import OrderedDict
from lightningcast.collect_data import read_netcdf
from DataGenerator import DataGeneratorPredictPatches

os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

LIST_OF_METRIC_FUNCTIONS = [
    keras_metrics.accuracy,
    keras_metrics.binary_accuracy,
    keras_metrics.binary_csi,
    keras_metrics.binary_frequency_bias,
    keras_metrics.binary_pod,
    keras_metrics.binary_pofd,
    keras_metrics.binary_peirce_score,
    keras_metrics.binary_success_ratio,
    keras_metrics.binary_focn,
]
METRIC_FUNCTION_DICT = {
    "accuracy": keras_metrics.accuracy,
    "binary_accuracy": keras_metrics.binary_accuracy,
    "binary_csi": keras_metrics.binary_csi,
    "binary_frequency_bias": keras_metrics.binary_frequency_bias,
    "binary_pod": keras_metrics.binary_pod,
    "binary_pofd": keras_metrics.binary_pofd,
    "binary_peirce_score": keras_metrics.binary_peirce_score,
    "binary_success_ratio": keras_metrics.binary_success_ratio,
    "binary_focn": keras_metrics.binary_focn,
}
#################################################################################################################################
from shapely import geometry
from geojson import Polygon, Feature, FeatureCollection
import geojson
from geojson_rewind import rewind
import json


def write_json(cs, outfile, re_upload=None, values=[0.25, 0.5, 0.75], empty=False):
    fc_list = []
    colors = ["0 0 255", "0 255 255", "255 0 255"]
    cter = 0

    if empty:
        fc_list = []
    else:
        for col in cs.collections:
            # Loop through all polygons that have the same intensity level
            print("len", len(col.get_paths()))
            for contour_path in col.get_paths():
                # Create the polygon for this intensity level
                # The first polygon in the path is the main one, the following ones are "holes"

                for ncp, cp in enumerate(contour_path.to_polygons()):
                    # print(cp)
                    try:
                        x = cp[:, 0]
                        y = cp[:, 1]
                    except TypeError:
                        # print('TypeError')
                        x = [a[0] for a in cp]
                        y = [a[1] for a in cp]
                    new_shape = geometry.Polygon(
                        [(round(i[0], 3), round(i[1], 3)) for i in zip(x, y)]
                    )
                    if ncp == 0:
                        poly = new_shape
                    else:
                        # Remove the holes if there are any
                        poly = poly.difference(new_shape)
                        # Can also be left out if you want to include all rings

                    # do something with polygon
                    poly = Feature(geometry=poly)
                    poly["properties"]["OPACITY"] = "0"
                    poly["properties"]["BOPACITY"] = "100"
                    poly["properties"]["COLOR"] = colors[cter]
                    poly["properties"]["BCOLOR"] = colors[cter]
                    poly["properties"]["WIDTH"] = "1"
                    poly["properties"]["INFO"] = (
                        "ICP: " + str(int(values[cter] * 100)) + "%"
                    )
                    if poly.is_valid:
                        fc_list.append(poly)
                    else:
                        print("poly not valid:")
                        print(poly)

            cter += 1

    fc = FeatureCollection(fc_list)
    fc_str = rewind(fc)  # makes json 'right-hand rule' compliant
    fc_str = json.dumps(fc_str, indent=2, sort_keys=True)  # pretty printing
    utils.mkdir_p(os.path.dirname(outfile))
    of = open(outfile, "w")
    of.write(fc_str)
    of.close()

    if (
        re_upload
    ):  # re_upload is either None or a string to full path of the re_upload script
        s = call([re_upload, outfile])


#################################################################################################################################
def find_latest(listened_file, sector):
    if not (os.path.isfile(listened_file)):
        print(listened_file + " is not a file or does not exist")
        sys.exit(1)
    f = open(listened_file, "r")
    lines = f.read().split("\n")
    print("Checking lines for " + sector)
    latest = None
    nlines = len(lines)
    for i in range(nlines - 1, 0, -1):
        if sector in lines[i]:
            latest = lines[i]
            break

    if latest:
        parts = latest.split("#")
        machine = parts[1]
        partition = "/" + machine.split(".")[0] + "_data"
        filestr = parts[-1]
        basefile = os.path.basename(filestr)
        fileparts = filestr.split("_")
        timestamp = fileparts[-1][1:14]
        dt = datetime.strptime(timestamp, "%Y%j%H%M%S")
        remoteDir = (os.path.dirname(filestr)).replace(
            "/data", partition
        )  # should be '/satbuf1_data', or '/satbuf2_data'
    else:
        basefile = remoteDir = dt = "None"

    return basefile, remoteDir, dt


#################################################################################################################################
def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx, array[idx]


#################################################################################################################################
def get_reps(reports_file):
    lines = [line.rstrip("\n") for line in open(reports_file)]
    reps = []
    init_dt = datetime.strptime("20" + reports_file[0:6], "%Y%m%d")
    for ll in lines:
        if "Time," in ll:
            continue
        parts = ll.split(",")  # parts[0] = HHMM, parts[-2] = lon, parts[-3] = lat
        hhmm = parts[0]
        td = timedelta(hours=int(hhmm[0:2]), minutes=int(hhmm[2:4]))
        if int(hhmm) >= 1200:
            dt = init_dt + td
        else:
            dt = init_dt + td + timedelta(hours=24)
        reps.append((dt, float(parts[-2]), float(parts[-3])))
    return reps


#################################################################################################################################
def predict_ltg(
    filelist,
    model,
    model_config_file,
    outdir=os.environ["PWD"] + "/OUTPUT/",
    hrrr_dir=pltg_data + "NWP/",
    glmdir=None,
    bounding_box=[-999],
    stride=4,
    input_dir=None,
    sandimg=False,
    alpha=0.25,
    values=[0.25, 0.5, 0.9],
    sector="RadC",
    savegrid=False,
    realtime=False,
    slot="GOES_East",
    re_upload=None,
    make_json=False,
):

    utils.mkdir_p(outdir)
    t00 = time.time()

    state_borders = cfeature.NaturalEarthFeature(
        category="cultural",
        name="admin_1_states_provinces_lakes",
        scale="50m",
        facecolor="none",
    )

    conv_model = load_model(model, custom_objects=METRIC_FUNCTION_DICT, compile=False)
    # n_inputs = len(conv_model.input_names)
    # print(conv_model.layers[0])
    # lay = conv_model.layers[0]
    # print(lay.input_shape)
    # print(lay.input_shape[1],lay.input_shape[2])

    mc = pickle.load(open(model_config_file, "rb"))
    svals = mc["scaling_values"]
    channels = mc["channels"]
    ninputs = len(mc["input_tuples"])
    scalar_vars = mc["scalar_vars"]
    inputs = mc["inputs"]

    # what HRRR data are we using?
    hrrr_vars = []
    all_hrrr_vars = [
        "mlcape",
        "u925",
        "v925",
        "u850",
        "v850",
        "u700",
        "v700",
        "u500",
        "v500",
    ]
    # patches
    for inp in inputs:
        for ch in inputs[inp]:
            if ch in all_hrrr_vars:
                hrrr_vars.append(ch)  # will add in correct order
    for sv in scalar_vars:
        if sv in all_hrrr_vars:
            hrrr_vars.append(sv)  # will add in correct order
    hrrr_uppercase = [k.upper() for k in hrrr_vars]

    # with respect to ABI CONUS IR grid #bounding_box=[SWlon, NElon, SWlat, NElat]
    # meso or CONUS? GE or GW?
    # get filenames
    if realtime:
        # slot is passed as an arg
        nlines = 1
        latest_filestr = filestr = "foo"
        sectortype = "conus" if (sector == "RadC") else "meso"
    else:
        with open(filelist) as fp:
            lines = [l.rstrip("\n") for l in fp.readlines()]
        fp.close()
        nlines = len(lines)
        files = lines[0].split()
        firstfile = files[0]
        sectortype = "conus" if ("RadC" in firstfile) else "meso"
        slot = "GOES_East" if ("_G16_" in firstfile) else "GOES_West"
    currhrrrfile = "foo"
    glmfile = None

    # static 2km lats, lons
    conus2km = Dataset(os.path.join(pltg, "src", "static", slot + ".nc"), "r")
    satheight = conus2km.variables["goes_imager_projection"].getncattr(
        "perspective_point_height"
    )
    lon0 = conus2km.variables["goes_imager_projection"].getncattr(
        "longitude_of_projection_origin"
    )
    sweep = conus2km.variables["goes_imager_projection"].getncattr("sweep_angle_axis")
    # projections
    p = Proj(proj="geos", h=satheight, lon_0=lon0, sweep=sweep)
    geoproj = ccrs.Geostationary(
        central_longitude=lon0, sweep_axis=sweep, satellite_height=satheight
    )

    # The projection x and y coords equals the scanning angle (in radians) multiplied by the satellite height (http://proj4.org/projections/geos.html)
    irx = conus2km.variables["x"][:] * satheight
    iry = conus2km.variables["y"][:] * satheight
    irxx, iryy = np.meshgrid(irx, iry)
    irlons, irlats = p(irxx, iryy, inverse=True)
    conus2km.close()

    # static 05km lats, lons
    conus05km = Dataset(os.path.join(pltg, "src", "static", slot + "_C02.nc"), "r")
    # The projection x and y coords equals the scanning angle (in radians) multiplied by the satellite height (http://proj4.org/projections/geos.html)
    visx = conus05km.variables["x"][:] * satheight
    visy = conus05km.variables["y"][:] * satheight
    visxx, visyy = np.meshgrid(visx, visy)
    vislons, vislats = p(visxx, visyy, inverse=True)
    conus05km.close()

    if bounding_box[0] != -999:
        crop = True
        xmin, ymin = p(bounding_box[0], bounding_box[2])
        xmax, ymax = p(bounding_box[1], bounding_box[3])

        xmin_ind, xmin = find_nearest(irx, xmin)
        xmax_ind, xmax = find_nearest(irx, xmax)
        ymin_ind, ymin = find_nearest(iry, ymin)
        ymax_ind, ymax = find_nearest(iry, ymax)

        startX = xmin_ind
        endX = xmax_ind
        startY = ymax_ind
        endY = ymin_ind
    else:
        # assume 'OCONUS' crop
        crop = True
        #    startX = 1510
        #    endX = 2499
        #    startY = 0
        #    endY = 970
        # assume 'CONUS' crop
        #    startX, endX = 55, 2300
        #    startY, endY = 32, 1040

        startX, endX = 1510, 1910
        startY, endY = 450, 850

        xmin = irx[startX]
        xmax = irx[endX]
        ymin = iry[endY]
        ymax = iry[
            startY
        ]  # N.B.: startY is north, where y is greater. That's why it uses ymax

    if crop:
        irlons = irlons[startY:endY, startX:endX]
        irlats = irlats[startY:endY, startX:endX]
        irxx = irxx[startY:endY, startX:endX]
        iryy = iryy[startY:endY, startX:endX]
        vislons = vislons[startY * 4 : endY * 4, startX * 4 : endX * 4]
        vislats = vislats[startY * 4 : endY * 4, startX * 4 : endX * 4]
        visxx = visxx[startY * 4 : endY * 4, startX * 4 : endX * 4]
        visyy = visyy[startY * 4 : endY * 4, startX * 4 : endX * 4]

    # half sizes
    hs = {
        "2km": int(mc["ps_2km"][0] / 2),
        "glm": int(mc["ps_2km"][0] / 2),
        "1km": int(mc["ps_1km"][0] / 2),
        "05km": int(mc["ps_05km"][0] / 2),
    }

    frames = {"05km": None, "1km": None, "2km": None, "glm": None}
    if len(mc["input_tuples"][0]) == 4:
        # we have a sequence, e.g., (7,32,32,1)
        hs_ind = 1
        nframes = mc["input_tuples"][0][0]  # 7, in the example above
    else:
        hs_ind = 0
        nframes = 1
    for i in range(ninputs):
        if mc["input_tuples"][i][hs_ind] == hs["05km"] * 2:
            if crop:
                frames["05km"] = np.empty(
                    (nframes, endY * 4 - startY * 4, endX * 4 - startX * 4)
                )
            else:  # assumes mesoscale sector
                frames["05km"] = np.empty((nframes, 2000, 2000))
        elif mc["input_tuples"][i][hs_ind] == hs["1km"] * 2:
            if crop:
                frames["1km"] = np.empty(
                    (nframes, endY * 2 - startY * 2, endX * 2 - startX * 2)
                )
            else:
                print("???")
        else:
            if crop:
                frames["2km"] = np.empty((nframes, endY - startY, endX - startX))
                frames["glm"] = np.empty((nframes, endY - startY, endX - startX))
            else:  # assumes mesoscale sector
                frames["2km"] = np.empty((nframes, 500, 500))
                frames["glm"] = np.empty((nframes, 500, 500))

    print("overhead:", time.time() - t00)

    # get filenames
    with open(filelist) as fp:
        lines = [l.rstrip("\n") for l in fp.readlines()]
    fp.close()
    nlines = len(lines)

    line_cter = 0
    frame_cter = 0
    while line_cter < nlines or realtime:

        if realtime:
            while latest_filestr == filestr or filestr == "None":
                filestr, remoteDir, satdt = find_latest(filelist, sector)
                if latest_filestr == filestr or filestr == "None":
                    print("Already processed:", latest_filestr)
                    time.sleep(10)
            # reset
            print("Beginning to process:", os.path.join(remoteDir, filestr))
            tstart = time.time()
            latest_filestr = filestr

        else:
            print(line_cter, lines[line_cter])

        pred_2km = None
        pred_1km = None
        pred_05km = None
        pred_glm = None
        pred_scalar = None
        while frame_cter < nframes:

            if realtime:
                # list of files for realtime
                visfile = glob.glob(
                    os.path.join(remoteDir, filestr.replace("??", "02"))
                )[0]
                irfile = glob.glob(
                    os.path.join(remoteDir, filestr.replace("??", "13"))
                )[0]
                files = [visfile, irfile]

                if "FED" in channels:
                    irbase = os.path.basename(irfile)
                    abidt = datetime.strptime(irbase.split("_s")[1][0:11], "%Y%j%H%M")
                    for pp in range(10):
                        newdt = abidt - timedelta(minutes=pp)
                        try:
                            glm_file = glob.glob(
                                newdt.strftime(
                                    os.path.join(
                                        glmdir,
                                        "%Y%m%d",
                                        "GLM",
                                        "goes_east",
                                        "agg",
                                        "%Y%m%d-%H%M*.netcdf.gz",
                                    )
                                )
                            )[0]
                        except IndexError:
                            # logging.warning
                            if pp == 9:
                                print(
                                    "Haven't found glm data in the last 10 mins in "
                                    + glmdir
                                )
                                print("sleeping 1 min...")
                                time.sleep(60)
                                noglm = True
                        else:
                            noglm = False
                            break
                    if noglm:
                        break  # should start at top of last while loop
                    files.append(os.path.basename(glm_file))

                if (
                    "mlcape" in scalar_vars
                ):  # FIXME hardcoded 'mlcape' to serve as flag for HRRR
                    # get hrrr remapped file
                    irbase = os.path.basename(irfile)
                    abidt = datetime.strptime(irbase.split("_s")[1][0:9], "%Y%j%H")
                    for ph in range(5):
                        newdt = abidt - timedelta(hours=ph)
                        try:
                            hrrr_file = glob.glob(
                                newdt.strftime(
                                    os.path.join(hrrr_dir, "%Y%m%d", "HRRR*%Y%m%d_%H*")
                                )
                            )[0]
                        except IndexError:
                            # logging.warning
                            # if(ph==4):
                            # use mean fill value?
                            pass
                        else:
                            break
                    files.append(hrrr_file)
            else:
                # archive mode
                files = lines[line_cter].split()

            input_list = (
                []
            )  # list of lists of grids to be passed into generator. Currently, this won't work with the time dimension or multiple channels per grid
            input_centers = (
                []
            )  # list of lists of centers (one per input, not one per channel)
            print(files)
            print("")
            for f in files:  # this is basically "for each channel" or "input"
                t0 = time.time()

                bf = os.path.basename(f)
                half_km = one_km = two_km = nwpdata = glm = False

                if bf.startswith("OR_ABI"):
                    try:
                        nc = Dataset(f, "r")
                    except OSError:
                        if (
                            any(ss in f for ss in ["/satbuf", "/arcdata/"])
                            or "/arcdata/" in input_dir
                        ):
                            # copy file locally, then delete.
                            # input_dir is: '/arcdata/goes/grb/goes16/'
                            parts = bf.split("_")
                            sector = parts[1].split("-")[2]
                            fts = parts[3][1:14]
                            dt_fts = datetime.strptime(fts, "%Y%j%H%M%S")
                            if not (realtime):
                                fullfile = dt_fts.strftime(
                                    os.path.join(
                                        input_dir,
                                        "%Y",
                                        "%Y_%m_%d_%j",
                                        "abi",
                                        "L1b",
                                        sector,
                                        f,
                                    )
                                )
                            localfile = os.path.join(os.environ["PWD"], f)
                            shutil.copyfile(fullfile, localfile)
                            nc = Dataset(localfile, "r")
                        else:
                            nc = Dataset(os.path.join(input_dir, f), "r")
                    radiance = nc.variables["Rad"][:]

                    basefile = os.path.basename(f)
                    baseparts = basefile.split("_")
                    channel = baseparts[1][-2:]
                    goes_position = nc.orbital_slot
                    scene_id = nc.scene_id
                    tce = datetime.strptime(
                        nc.time_coverage_end[0:16], "%Y-%m-%dT%H:%M"
                    )

                    if "C02_G" in f:
                        idx = "CH02"
                        # convert vis radiance to reflectance
                        result = radiance * nc.variables["kappa0"][:]
                        visdata = np.copy(result)
                        if crop:
                            visdata = visdata[
                                startY * 4 : endY * 4, startX * 4 : endX * 4
                            ]
                        fac = 4
                        ts = stride * 4
                        vis_stride = ts
                        hpy, hpx = int(mc["ps_05km"][0] / 2.0), int(
                            mc["ps_05km"][1] / 2.0
                        )
                        vis_hpy, vis_hpx = hpy, hpx
                        vis_ny, vis_nx = result.shape
                        half_km = True
                    elif "C01_G" in f or "C03_G" in f or "C05_G" in f:
                        idx = "CH05"
                        fac = 2
                        ts = stride * 2
                        result = radiance * nc.variables["kappa0"][:]
                        one_km = True
                        hpy, hpx = [int(v / 2) for v in meta["ps"]["1km"]]
                    else:
                        idx = "CH13"
                        # convert IR radiance to brightness temperature
                        fac = 1
                        result = (
                            nc.variables["planck_fk2"][:]
                            / (np.log((nc.variables["planck_fk1"][:] / radiance) + 1))
                            - nc.variables["planck_bc1"][:]
                        ) / nc.variables["planck_bc2"][:]
                        irdata = np.copy(result)
                        if crop:
                            irdata = irdata[startY:endY, startX:endX]
                        ts = stride
                        hpy, hpx = int(mc["ps_2km"][0] / 2.0), int(
                            mc["ps_2km"][1] / 2.0
                        )
                        two_km = True
                        timestamp = getattr(nc, "time_coverage_start")[0:16]
                        dt = datetime.strptime(timestamp, "%Y-%m-%dT%H:%M")

                elif bf.startswith("HRRR"):
                    if currhrrrfile != bf:
                        hrrrdict = OrderedDict()
                        for k in hrrr_vars:
                            hrrrdict[k] = -1  # initialize
                        hstat = read_netcdf(f, data=hrrrdict)
                        currhrrrfile = bf  # reset
                        scalehrrr = True  # we only scale if new data
                    else:
                        scalehrrr = False
                    nwpdata = True
                    fac = 1
                    ts = stride
                else:  # glm file
                    idx = "FED"
                    glm = True
                    fac = 1
                    ts = stride
                    hpy, hpx = int(mc["ps_2km"][0] / 2.0), int(mc["ps_2km"][1] / 2.0)
                    two_km = True

                    glmdate = f[0:8]
                    glmfile = os.path.join(
                        glmdir, glmdate, "GLM", "goes_east", "agg", f
                    )
                    localfile = os.path.join(os.environ["PWD"], f)
                    shutil.copyfile(glmfile, localfile)
                    if f.endswith("gz"):
                        s = call(["gunzip", "-f", f])
                        nc = Dataset(f[:-3], "r")  # assumes we prepend input_dir
                        result = nc.variables["flash_extent_density"][:]
                        s = call(["gzip", "-f", f[:-3]])
                    else:
                        nc = Dataset(f, "r")  # assumes we prepend input_dir
                        result = nc.variables["flash_extent_density"][:]
                    result = np.ma.filled(
                        result
                    )  # unfill all of those masked pixels with 0s

                try:
                    nc.close()
                except RuntimeError:
                    pass

                try:
                    os.remove(localfile)
                except (NameError, IOError, OSError):
                    pass

                if nwpdata:
                    if crop and scalehrrr:
                        for k in hrrrdict:
                            hrrrdict[k] = hrrrdict[k][
                                startY * fac : endY * fac, startX * fac : endX * fac
                            ]
                            hrrrdict[k] = (
                                hrrrdict[k] - svals.loc[k, "mean"]
                            ) / svals.loc[k, "std"]
                    elif scalehrrr:
                        for k in hrrrdict:
                            hrrrdict[k] = (
                                hrrrdict[k] - svals.loc[k, "mean"]
                            ) / svals.loc[k, "std"]
                    ny, nx = hrrrdict[
                        hrrr_vars[-1]
                    ].shape  # assumes all hrrr_vars are the same shape

                    # discern if patches
                    hrrr_input_list = []
                    for k in hrrrdict:
                        for inp in inputs:
                            if k == inputs[inp]:
                                hrrr_input_list.append(
                                    hrrrdict[k]
                                )  # will add in correct order
                    if len(hrrr_input_list) > 0:
                        input_list.append(hrrr_input_list)
                else:
                    if crop:
                        chopped_grid = result[
                            startY * fac : endY * fac, startX * fac : endX * fac
                        ]
                    else:
                        chopped_grid = result
                    ny, nx = chopped_grid.shape

                    # scale the data
                    chopped_grid = (chopped_grid - svals.loc[idx, "mean"]) / svals.loc[
                        idx, "std"
                    ]
                    if half_km:
                        frames["05km"][frame_cter] = chopped_grid
                    elif one_km:
                        frames["1km"][frame_cter] = chopped_grid
                    elif two_km and not (glm):
                        frames["2km"][frame_cter] = chopped_grid
                    else:  # glm #FIXME not tested
                        frames["glm"][frame_cter] = chopped_grid

                    input_list.append([chopped_grid])

                if half_km:
                    k = "05km"
                    xvec = np.arange(nx)
                    yvec = np.arange(ny)
                    xgrid, ygrid = np.meshgrid(xvec, yvec)
                    xcenters = xgrid[hs[k] : -hs[k] : ts, hs[k] : -hs[k] : ts].ravel()
                    ycenters = ygrid[hs[k] : -hs[k] : ts, hs[k] : -hs[k] : ts].ravel()
                    centers = [x for x in zip(ycenters, xcenters)]
                    viscenters = centers
                    input_centers.append(centers)

                if one_km:
                    k = "1km"
                    xvec = np.arange(nx)
                    yvec = np.arange(ny)
                    xgrid, ygrid = np.meshgrid(xvec, yvec)
                    xcenters = xgrid[hs[k] : -hs[k] : ts, hs[k] : -hs[k] : ts].ravel()
                    ycenters = ygrid[hs[k] : -hs[k] : ts, hs[k] : -hs[k] : ts].ravel()
                    centers = [x for x in zip(ycenters, xcenters)]
                    input_centers.append(centers)

                if two_km or nwpdata:
                    if glm:
                        k = "glm"
                    else:
                        k = "2km"
                    xvec = np.arange(nx)
                    yvec = np.arange(ny)
                    xgrid, ygrid = np.meshgrid(xvec, yvec)
                    xcenters = xgrid[hs[k] : -hs[k] : ts, hs[k] : -hs[k] : ts].ravel()
                    ycenters = ygrid[hs[k] : -hs[k] : ts, hs[k] : -hs[k] : ts].ravel()
                    centers = [x for x in zip(ycenters, xcenters)]

                    if (
                        nwpdata
                    ):  # note that HRRR uses 2km input centers, but a different half size
                        if len(hrrr_input_list) > 0:
                            input_centers.append(centers)
                    else:
                        input_centers.append(centers)

                    if f == files[-1]:  # 'last input'
                        # do scalars within two_km block
                        if len(scalar_vars):
                            clats = [irlats[c[0], c[1]] for c in centers]
                            clons = [irlons[c[0], c[1]] for c in centers]
                            for sv in scalar_vars:
                                svu = sv.upper()
                                if svu == "SOLZEN":
                                    dts = [dt] * len(centers)
                                    dat, solaz = solar_angles(dts, lat=clats, lon=clons)
                                elif svu == "SATZEN":
                                    dat = sat_zen_angle(
                                        xlat=clats, xlon=clons, satlon=lon0
                                    )
                                elif svu == "LAT":
                                    dat = np.copy(clats)
                                elif svu == "LON":
                                    dat = np.copy(clons)
                                elif svu in hrrr_uppercase:
                                    dat = np.array(
                                        [
                                            hrrrdict[svu.lower()][c[0], c[1]]
                                            for c in centers
                                        ]
                                    )  # usu, nmd fct is what makes it an array
                                else:
                                    print("Can't find " + sv + " in scalar_vars!")
                                    sys.exit(1)

                                # hrrr data has already been scaled
                                if not nwpdata:
                                    dat, _ = nmd(
                                        dat,
                                        scaling_values=svals,
                                        index=sv.lower(),
                                        test=True,
                                    )
                                if pred_scalar is None:
                                    pred_scalar = dat + 0.0
                                else:
                                    pred_scalar = np.vstack((pred_scalar, dat))
                            pred_scalar = np.transpose(pred_scalar)

            # end for each file in one timestep
            frame_cter += (
                1  # move on to next 'frame' (if using temporal component in model)
            )
            line_cter += 1
        # end inner while

        print("making predictions...")
        t0 = time.time()

        try:
            gen = DataGeneratorPredictPatches(
                input_list,
                mc["input_tuples"],
                input_centers,
                scalars=pred_scalar,
                batch_size=32,
            )
            new_preds = conv_model.predict(
                gen, workers=0
            )  # no multiprocessing...but does multithreading by default
        except ValueError as err:
            print(
                err
            )  # in realtime, this is an error possibly caused by certain input data not being available
            # will just move on to next scene
        else:

            print("predict time:", "{:.3f}".format(time.time() - t0))
            tmp_preds = np.zeros((len(input_centers[0])))
            tmp_preds[0 : len(new_preds)] = new_preds[
                :, 0
            ]  # by using a generator, we are probably a few predictions short
            new_preds = tmp_preds
            new_preds = new_preds.reshape(
                int(np.ceil((ny - hs["2km"] * 2) / stride)),
                int(np.ceil((nx - hs["2km"] * 2) / stride)),
            )

            if realtime or make_json:
                print("making contours and json...")
                tj = time.time()
                fig = plt.figure(figsize=(15, 15))
                ax = plt.axes(projection=geoproj)
                # cxx = np.array([visxx[c[0],c[1]] for c in viscenters]).reshape(new_preds.shape)
                # cyy = np.array([visyy[c[0],c[1]] for c in viscenters]).reshape(new_preds.shape)
                lons = np.array([vislons[c[0], c[1]] for c in viscenters]).reshape(
                    new_preds.shape
                )
                lats = np.array([vislats[c[0], c[1]] for c in viscenters]).reshape(
                    new_preds.shape
                )

                # error fixing
                new_preds = np.ma.fix_invalid(new_preds, fill_value=0)
                new_preds[lons < -130] = 0
                print("max new_preds:", np.max(new_preds))

                contour_img = ax.contour(
                    lons, lats, new_preds, values, transform=ccrs.PlateCarree()
                )
                print(len(contour_img.collections[0].get_paths()))
                outjson = os.path.join(
                    outdir,
                    dt.strftime("%Y%m%d"),
                    "PLTG",
                    dt.strftime("PLTG_%Y%m%d_%H%M00.force.json"),
                )
                empty = False if (np.max(new_preds) > values[0]) else True
                write_json(contour_img, outjson, re_upload=re_upload, empty=empty)
                print("wrote out ", outjson)
                if realtime:
                    print("total time:", "{:.3f}".format(time.time() - tstart))

            elif savegrid:
                out_x = irx[hs["2km"] : -hs["2km"] : stride] / satheight
                out_y = iry[hs["2km"] : -hs["2km"] : stride] / satheight

                # error fixing
                new_preds = np.ma.fix_invalid(new_preds, fill_value=0)
                lons = np.array([vislons[c[0], c[1]] for c in viscenters]).reshape(
                    new_preds.shape
                )
                new_preds[lons < -130] = 0

                dims = OrderedDict()
                dims["x"] = len(out_x)
                dims["y"] = len(out_y)
                dims["g"] = 1

                dataset = {}
                dataset["x"] = {
                    "data": out_x,
                    "dims": ("x",),
                    "atts": {
                        "units": "rad",
                        "axis": "X",
                        "long_name": "GOES fixed grid projection x-coordinate",
                        "standard_name": "projection_x_coordinate",
                    },
                }
                dataset["y"] = {
                    "data": out_y,
                    "dims": ("y",),
                    "atts": {
                        "units": "rad",
                        "axis": "Y",
                        "long_name": "GOES fixed grid projection y-coordinate",
                        "standard_name": "projection_y_coordinate",
                    },
                }
                dataset["goes_imager_projection"] = {
                    "data": np.int32(1),
                    "dims": ("g",),
                    "atts": {},
                }
                conus2km = Dataset(
                    os.path.join(pltg, "src", "static", slot + ".nc"), "r"
                )
                for att in conus2km.variables["goes_imager_projection"].ncattrs():
                    dataset["goes_imager_projection"]["atts"][att] = conus2km.variables[
                        "goes_imager_projection"
                    ].getncattr(att)
                conus2km.close()
                dataset["intense_convection_probability"] = {
                    "data": new_preds,
                    "dims": ("y", "x"),
                    "atts": {"model": model},
                }
                atts = OrderedDict()
                atts["orbital_slot"] = slot
                atts["sector"] = sector
                atts["datetime"] = dt.strftime("%Y%m%d-%H%M%S UTC")

                netcdf = os.path.join(
                    outdir, dt.strftime("%Y%m%d/PLTG/PLTG_%Y%m%d-%H%M%S.netcdf")
                )
                status = utils.write_netcdf(
                    netcdf, dataset, dims, atts=atts, wait=True, gzip=False
                )

            else:  # make an image

                # get aspect ratio for IR data
                default_imgsize = 10  # inches
                x_range = endX - startX
                y_range = endY - startY
                asp = y_range / x_range
                if asp > 1:
                    fig_width_in = int(default_imgsize * asp ** (-1))
                    fig_height_in = int(default_imgsize)
                else:
                    fig_width_in = int(default_imgsize)
                    fig_height_in = int(default_imgsize * asp)

                if glmdir:
                    # find best file
                    glmfile = glob.glob(
                        tce.strftime(
                            os.path.join(
                                glmdir,
                                "%Y%m%d",
                                "GLM",
                                "goes_east",
                                "agg",
                                "%Y%m%d-%H%M*.netcdf*",
                            )
                        )
                    )
                    if not (glmfile):
                        for mm in [1, 2, 3, 4, 5]:
                            newdt = tce - timedelta(minutes=mm)
                            glmfile = glob.glob(
                                newdt.strftime(
                                    os.path.join(
                                        glmdir,
                                        "%Y%m%d",
                                        "GLM",
                                        "goes_east",
                                        "agg",
                                        "%Y%m%d-%H%M*.netcdf*",
                                    )
                                )
                            )
                            if glmfile:
                                break
                    if glmfile:
                        glmdata = {"flash_extent_density": -1}
                        ss = read_netcdf(glmfile[0], data=glmdata)
                        imgdata = glmdata["flash_extent_density"][
                            startY:endY, startX:endX
                        ]
                        vmin = 0
                        vmax = 80
                        inc = 10
                        cmap = plt.get_cmap("autumn_r")
                        cbar_label = "GLM flash extent density [fl/(5-min)]"
                        out_prefix = "GLMFED"
                else:
                    # assume an IR image
                    imgdata = frames["2km"][-1]
                    vmin = 180
                    vmax = 310
                    inc = 10
                    cmap = irw_ctc()
                    cbar_label = "Channel " + channel + " brightness temperature [K]"
                    out_prefix = "C" + channel

                fig = plt.figure(figsize=((fig_width_in, fig_height_in)))
                ax = fig.add_axes([0, 0, 1, 1], projection=geoproj)
                # set extent of image
                ax.set_extent([xmin, xmax, ymin, ymax], crs=geoproj)
                ax.add_feature(cfeature.BORDERS)
                ax.add_feature(state_borders, edgecolor="tan")

                #  pcimg = ax.pcolormesh(visxx, visyy, np.sqrt(visdata), vmin=0, vmax=1, transform=geoproj, cmap=plt.get_cmap('Greys_r'))
                #  imgd = ma.masked_less(zoom(imgdata,4,order=1),1); irxz = zoom(irxx,4,order=1); iryz = zoom(iryy,4,order=1)
                #  pcimg = ax.pcolormesh(irxz, iryz, imgd, vmin=vmin, vmax=vmax, transform=geoproj, cmap=cmap, alpha=0.65, linewidth=0, antialiased=True)

                irimg = ax.pcolormesh(
                    irxx, iryy, irdata, vmin=180, vmax=310, cmap=plt.get_cmap("Greys")
                )
                cbaxes = fig.add_axes(
                    [0.01, -0.05, 0.27, 0.03]
                )  # left bottom width height
                cbar = plt.colorbar(
                    irimg,
                    orientation="horizontal",
                    pad=0.01,
                    extendfrac=0,
                    drawedges=0,
                    ax=ax,
                    ticks=np.arange(180, 301, 40),
                    cax=cbaxes,
                )
                cbar.set_label(
                    "ABI 10.35-$\mu$m brightness temperature [K]", fontsize=10
                )

                glmimg = ax.pcolormesh(
                    irxx,
                    iryy,
                    ma.masked_less(imgdata, 0.1),
                    vmin=vmin,
                    vmax=vmax,
                    cmap=plt.get_cmap("autumn_r"),
                    linewidth=0,
                    antialiased=True,
                    alpha=0.5,
                )
                cbaxes2 = fig.add_axes(
                    [0.31, -0.05, 0.27, 0.03]
                )  # left bottom width height
                cbar2 = plt.colorbar(
                    glmimg,
                    orientation="horizontal",
                    pad=0.01,
                    extendfrac=0,
                    drawedges=0,
                    ax=ax,
                    ticks=np.arange(10, 71, 20),
                    cax=cbaxes2,
                )
                cbar2.set_label("Max GLM FED [fl (5 min)$^{-1}$]", fontsize=10)
                cbar2.set_alpha(1.0)
                cbar2.draw_all()

                # contours
                cxx = np.array([irxx[c[0], c[1]] for c in centers]).reshape(
                    new_preds.shape
                )
                cyy = np.array([iryy[c[0], c[1]] for c in centers]).reshape(
                    new_preds.shape
                )
                colors = ["#0000FF", "#00FFFF", "#00FF00", "#CC00FF"]
                values = [0.1, 0.25, 0.5, 0.75]
                linewidths = [0.5, 0.75, 0.75, 1]
                labels = [str(int(v * 100)) + "%" for v in values]
                contour_img = ax.contour(
                    cxx, cyy, new_preds, values, colors=colors, linewidths=linewidths
                )  # , cmap=plt.get_cmap('cividis'))
                custom_lines = [
                    Line2D([0], [0], color=c, lw=lw * 2)
                    for c, lw in zip(colors, linewidths)
                ]
                leg2 = ax.legend(
                    custom_lines,
                    labels,
                    fontsize=11,
                    loc="lower right",
                    handlelength=1.5,
                    ncol=len(labels),
                    bbox_to_anchor=(1, -0.06),
                )
                ctext = plt.text(
                    0.65,
                    -0.09,
                    "Probability of lightning in 30 min",
                    fontsize=14,
                    transform=ax.transAxes,
                )

                ax.set_title(
                    (" ").join([goes_position, scene_id, timestamp + " UTC"]),
                    fontsize=12,
                )

                fig.savefig(
                    os.path.join(
                        outdir,
                        ("_").join(
                            [out_prefix, scene_id, timestamp.replace(":", "") + ".png"]
                        ),
                    ),
                    bbox_inches="tight",
                )
                plt.close()

        # back up frame and frame_cter
        frame_cter -= 1
        if nframes > 1:
            for k in frames:
                if isinstance(frames[k], np.ndarray):
                    frames[k][0 : nframes - 1] = frames[k][1:nframes]


################################################################################################################################################################################
if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="Takes a list of file sequence or rabbitMQ real-time list of GOES data and \
                                                will make predictions of lightning in the next 30 min. Can make images, jsons, or grids."
    )
    parser.add_argument(
        "filelist",
        help="A list of channel filenames. Channels should be in order that model expects them. See chan_filelist.txt for example.\
                                 If realtime is set, then this argument is the listened log output from amqpfind.",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "model", help="CNN model used for predictions.", type=str, nargs=1
    )
    parser.add_argument(
        "model_config_file",
        help="This file contains mean and SD of channels (a pickle file).",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-i",
        "--input_dir",
        help='prepend this to each file in filelist. If "/arcdata/....", will copy files locally then remove',
        type=str,
        nargs=1,
        default=[None],
    )
    parser.add_argument(
        "-pv",
        "--prob_vals",
        help="probability values to contour (0 to 1). Default = 0.25 0.5 0.9",
        type=float,
        nargs=3,
        default=[0.25, 0.5, 0.9],
    )
    parser.add_argument(
        "-s",
        "--stride",
        help="moving stride size in both directions (units=pixels). \
                                            This is for the 2km channels. The 1km and 0.5km channels will \
                                            multiply this by 2 and 4, respectively. Default=4.",
        default=[4],
        type=int,
        nargs=1,
    )
    parser.add_argument(
        "-hd",
        "--hrrr_dir",
        help="root NWP/HRRR dir with grids remapped to ABI projection. Default=$PLTG_DATA/NWP/",
        type=str,
        nargs=1,
        default=[pltg_data + "NWP/"],
    )
    parser.add_argument(
        "-g",
        "--glmdir",
        help="root glm dir. Default: /apollo/grain/saved_probsevere_data/lightning/ ...will append yyyymmdd/GLM/goes_east/agg/",
        type=str,
        nargs=1,
        default=["/apollo/grain/saved_probsevere_data/lightning/"],
    )
    parser.add_argument(
        "-b",
        "--bounding_box",
        help="Bounding box of image. SWlon, NElon, SWlat, NElat. Default is entire image.",
        default=[-999],
        type=float,
        nargs=4,
    )
    parser.add_argument(
        "-o",
        "--outdir",
        help="Root output directory. Default=$PWD/OUTPUT/",
        default=[os.path.join(os.environ["PWD"], "OUTPUT/")],
        nargs=1,
        type=str,
    )
    parser.add_argument(
        "-a",
        "--alpha",
        help="alpha for IR image in sandwich. Default is 0.25",
        default=[0.25],
        type=float,
        nargs=1,
    )
    parser.add_argument(
        "-sec",
        "--sector",
        help="RadC, RadM1, or RadM2. Use only if realtime=True",
        default=["RadC"],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-sl",
        "--slot",
        help="orbital slot...GOES_East or GOES_West. Only used if realtime=True",
        default=["GOES_East"],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-ru",
        "--re_upload",
        help="Full path to re_upload executable. Will execute after json is written.",
        default=[None],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-sg",
        "--savegrid",
        help="Save a netcdf of the ICP to outdir. Only in offline mode right now.",
        action="store_true",
    )
    parser.add_argument(
        "-mj",
        "--make_json",
        help="Save a json to outdir. --realtime does this by default.",
        action="store_true",
    )
    parser.add_argument(
        "-r",
        "--realtime",
        help="Set for realtime processing. filelist arg is the watched file from ampqfind.",
        action="store_true",
    )

    args = parser.parse_args()

    predict_ltg(
        args.filelist[0],
        args.model[0],
        args.model_config_file[0],
        outdir=args.outdir[0],
        stride=args.stride[0],
        alpha=args.alpha[0],
        bounding_box=args.bounding_box,
        input_dir=args.input_dir[0],
        values=args.prob_vals,
        glmdir=args.glmdir[0],
        hrrr_dir=args.hrrr_dir[0],
        sector=args.sector[0],
        slot=args.slot[0],
        realtime=args.realtime,
        re_upload=args.re_upload[0],
        savegrid=args.savegrid,
        make_json=args.make_json,
    )
