import os, sys
from datetime import datetime, timedelta
import logging
from logging.config import dictConfig

import inspect

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.insert(0, os.path.dirname(current_dir))

from lightningcast.logging_def import *
import numpy as np
import time
import re
import pandas as pd
import urllib

# This program runs infinitely and records new on-demand DSS events to a CSV file.

# Logging info
logDict = logging_def(logfile=None)
# do this regardless of realtime/offline, so we get formatted stdout messages
dictConfig(logDict)
# send stderr to logger
stderr_logger = logging.getLogger("STDERR")
sl = StreamToLogger(stderr_logger, logging.ERROR)

logging.info("Process started")

# config info
pltg_data = os.environ["PLTG_DATA"]
remote_csv = os.environ["ONDEMAND_CSV"]
dss_duration_thresh = int(os.environ["DSS_DURATION_THRESH"])
dss_save_file = f"{pltg_data}/dss_record.csv"

while True:
    logging.info("Checking for new DSS events...")

    try:
        dss_events = pd.read_csv(remote_csv)

        ts_of_request = list(dss_events["Timestamp"])  # type=str
        emails = list(dss_events["Email Address"])  # type=str
        event_name = list(dss_events["Unique name of event"])  # type=str
        lats = list(dss_events["Latitude of event (degrees north)"])  # type=float
        lons = list(dss_events["Longitude of event (degrees west)"])  # type=float
        begin_event_ts = list(
            dss_events["Beginning date & time of request (UTC)"]
        )  # type=str
        end_event_ts = list(dss_events["End date & time of request (UTC)"])  # type=str
    except urllib.error.HTTPError as err:
        logging.error(str(err))
    except KeyError as err:
        logging.critical("Bad key names in on-demand CSV")
        logging.critical(err)
        sys.exit(1)
    else:

        if os.path.isfile(dss_save_file):
            # Get the saved events
            of = open(dss_save_file, "r")
            saved_events = of.readlines()[1:]  # skip header
            of.close()
            write_headers = False
        else:
            write_headers = True

        # Open file for writing
        of = open(dss_save_file, "a")
        if write_headers:
            of.write(",".join(list(dss_events)) + "\n")

        for ii in range(len(emails)):

            # First check to see if we've already recorded this event
            record = (
                ",".join(
                    [
                        ts_of_request[ii],
                        emails[ii],
                        event_name[ii],
                        str(lats[ii]),
                        str(lons[ii]),
                        begin_event_ts[ii],
                        end_event_ts[ii],
                    ]
                )
                + "\n"
            )
            if record in saved_events:
                continue

            # The next four checks are just in case the Google Apps Script hasn't deleted the line/event/record yet
            # Check for valid email
            if not (
                bool(re.match(r"^(probsevere.*|.*noaa\.gov|.*wisc\.edu)+$", emails[ii]))
            ):
                logging.info(f"Invalid email: {emails[ii]}")
                continue

            # Check for redundant event name that occurs earlier
            if ii > 0 and event_name[ii] in event_name[0:ii]:
                logging.info(f"Redundant event name: {event_name[ii]}")
                continue

            # Check event_name - must be alphanumeric, underscores, or hyphens only
            if not (bool(re.match(r"^[A-Za-z0-9_\- ]+$", event_name[ii]))):
                logging.info(f"Bad characters: {event_name[ii]}")
                continue

            begin_dt = datetime.strptime(
                begin_event_ts[ii], "%m/%d/%Y %H:%M:%S"
            )  # assumed to be UTC
            end_dt = datetime.strptime(
                end_event_ts[ii], "%m/%d/%Y %H:%M:%S"
            )  # assumed to be UTC

            # Check that end_dt is after begin_dt
            # Also check that duration is within bounds.
            if (
                end_dt <= begin_dt
                or (end_dt - begin_dt).seconds / 86400 > dss_duration_thresh
            ):
                logging.info(
                    f"Bad times or too long of duration ({begin_event_ts[ii]} // {end_event_ts[ii]}"
                )
                continue  # 7-day duration limit

            # Passed all of our checks, so record the event
            record = ",".join(
                [
                    ts_of_request[ii],
                    emails[ii],
                    event_name[ii],
                    str(lats[ii]),
                    str(lons[ii]),
                    begin_event_ts[ii],
                    end_event_ts[ii],
                ]
            )
            of.write(record + "\n")
            logging.info(f"Recorded event: {record}")

        of.close()

    logging.info("Sleeping...")
    time.sleep(1800)
